/** 
 *   Camera Control for STAR I.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdarg.h>
#include <unistd.h>
#include <malloc.h>
#include <gpib/ib.h>
#include "camera_ctrl.h"
#include "my_stdio.h"

static int gpib_p_addr = -1;


#define MY_IBWRT_OLD(dev, fmt, arg...)\
        if ( my_ibwrt(dev, fmt, ## arg) & ERR ){ return -1; } 

#define MY_IBWRT(dev, fmt, arg...)\
        if ( my_ibwrite_generic(ibwrt, dev, fmt, ## arg) & ERR ){return -1;} 

#define MY_IBWRTA(dev, fmt, arg...)\
        if ( my_ibwrite_generic(ibwrta, dev, fmt, ## arg) & ERR ){return -1;} 

/*
#define  PUT_TITLE(dev)\
  	MY_IBWRT(dev,"PU;PA 200,205;TEXT|-- SKA Survey (IAR) --|;");
*/

static int my_ibwrite_generic(
	int (*write_func)(int ud, const void *data, long num_bytes),
	int dev, const char *fmt, ...)
{
	va_list ap;
	int l = -1;
	int n, size = 80;
	char *buff = NULL;

	if ((buff = malloc (size)) == NULL) return l;
	while (1) {
		/* Try to print in the allocated space. */
		va_start(ap, fmt);
		n = vsnprintf (buff, size, fmt, ap);
		va_end(ap);
		if (n > -1 && n < size) {
			break;
		}
		else {
			/* Else try again with more space. */
			if (n > -1)    /* glibc 2.1 */
				size = n+1; /* precisely what is needed */
			else           /* glibc 2.0 */
				size *= 2;  /* twice the old size */
			if ((buff = realloc (buff, size)) == NULL)
				return l;
		}
	}

	if (buff != NULL) {
		PVERB(PINFO,"# Send: %s\n", buff);
#ifdef DEMO_MODE
		l = write(dev, buff, n);
		write(dev, "\n", 1);

#else
		l = write_func(dev, buff, n);
#endif

#ifdef VERBOSE_MODE

		if (l & ERR && !(l & TIMO)) {
			SYSLOG(LOG_ERR,"Error sending: %s [ibstat: 0x%x]\n", 
			      buff, l);
		}
#endif
		free(buff);
	}

	return l;
}


int set_gpib_controller(int gpad)
{
	gpib_p_addr = gpad;
	/* Sets GPIB primary address */
	ibconfig(gpad, IbcPAD, 0);
	/* Sets GPIB secondary address */
	ibconfig(gpad, IbcSAD, 0);
	/* Sets timeout for io operations */
	ibconfig(gpad, IbcTMO, T30ms);
	/* If setting is nonzero, EOI is asserted with last byte on
           writes */
	ibconfig(gpad, IbcEOT, 1);
  	/* Automatic serial polling is disable */
	ibconfig(gpad, IbcAUTOPOLL, 0);
	ibconfig(gpad, IbcSC, 0);
	ibconfig(gpad, IbcEOSrd, 0);
	ibconfig(gpad, IbcEOSwrt, 0);
	ibconfig(gpad, IbcEOScmp, 0);
	ibconfig(gpad, IbcEOSchar, 0x00);
	ibconfig(gpad, IbcTIMING, T1_DELAY_350ns); /* FIXME */

	return 0;
}



int wait_camera_controller_on(int pad)
{
	short listen;
	int done = 0;

	do{
		/* check if listener is present */
		ibln(gpib_p_addr, pad, NO_SAD, &listen );
		if(ibsta & ERR){
			if(iberr == EBUS){
				SYSLOG(LOG_ERR,
				       "Camera controller not found in primary address %d",
				       pad);
			}else{
				SYSLOG(LOG_ERR,"Camera: init error [0x%x]", iberr);
			}
		}
		if( ! listen ){
			SYSLOG(LOG_ERR,
			       "Camera controller not found in primary address %d",
			       pad);
		}else{
			done = 1;
		}
	}while(!done);

	return 0;
}


int open_camera_controller(int pad)
{
	int dev;

#ifdef DEMO_MODE
	return 1;
#endif

	dev = ibdev(gpib_p_addr, pad, 0, CC_TIME_OUT, 1, 0);
	if (dev < 0) {
	  return dev;
	}
	wait_camera_controller_on(pad);
	/* Sets timeout for io operations */
	ibconfig(pad, IbcTMO, T30ms);
	/* If setting is nonzero, EOI is asserted with last byte on
           writes */
	ibconfig(pad, IbcEOT, 1);
	ibconfig(pad, IbcEOSrd, 0);
	ibconfig(pad, IbcEOSwrt, 0);
	ibconfig(pad, IbcEOScmp, 0);
	ibconfig(pad, IbcEOSchar, 0x00);

	return dev;
}



/******************************************************************************
 * Camera controller commands:
 *
 **/



static int get_int_from_cc(int dev, char *cmd, int *val)
{
	char ctmp[RW_BUF_SIZE];
	u_char spr;

	if ( my_ibwrite_generic(ibwrt, dev, "@%c", cmd) & ERR) 
		return -1;
	while (!(ibrsp(dev, &spr) & ERR)) {
		if (spr == CMD_OK) {
			break;
		}else if (spr == CMD_ERROR) {
			return -1;
		}
	}
	if (ibrd(dev, ctmp, RW_BUF_SIZE) & ERR) {
		return -1;
	}else{
		*val = atoi(ctmp);

	}		

	return 0;
}


static int get_str_from_cc(int dev, char *cmd, char *str, int len)
{
	u_char spr;

	if ( my_ibwrite_generic(ibwrt, dev, "@%c", cmd) & ERR) 
		return -1;
	while (!(ibrsp(dev, &spr) & ERR)) {
		if (spr == CMD_OK) {
			break;
		}else if (spr == CMD_ERROR) {
			return -1;
		}
	}
	if (ibrd(dev, str, len) & ERR) {
		str = NULL;
		return -1;
	}

	return 0;
}


static int put_int_to_cc(int dev, char *cmd, int val)
{
	u_char spr;

	if ( my_ibwrite_generic(ibwrt, dev, "!%c%d", cmd, val) & ERR) 
		return -1;
	while (!(ibrsp(dev, &spr) & ERR)) {
		if (spr == CMD_OK) {
			break;
		}else if (spr == CMD_ERROR) {
			return -1;
		}
	}

	return 0;
}


static int put_str_to_cc(int dev, char *cmd, char *str)
{
	u_char spr;

	if ( my_ibwrite_generic(ibwrt, dev, "!%c%s", cmd, str) & ERR) 
		return -1;
	while (!(ibrsp(dev, &spr) & ERR)) {
		if (spr == CMD_OK) {
			break;
		}else if (spr == CMD_ERROR) {
			return -1;
		}
	}

	return 0;
}


int get_exposure_time(int dev, int *time)
{
	return get_int_from_cc(dev, "A", time);
}
		

int set_exposure_time(int dev, int time)
{
	return put_int_to_cc(dev, "A", time);
}


int get_gain(int dev, int *gain)
{
	return get_int_from_cc(dev, "B", gain);
}

		
int set_gain(int dev, int gain)
{
	if (gain == GAIN_HIGH || gain == GAIN_LOW)
		return put_int_to_cc(dev, "B", gain);
	else
		return -1;
}


int get_temp_regulation(int dev, int *temp)
{
	return get_int_from_cc(dev, "C", temp);
}

		
int set_temp_regulation(int dev, int temp)
{
	if (temp == TEMP_REG_ON || temp == TEMP_REG_OFF)
		return put_int_to_cc(dev, "C", temp);
	else
		return -1;
}


int get_box_y0(int dev, int *y)
{
	return get_int_from_cc(dev, "D", y);
}
		

int set_box_y0(int dev, int y)
{
	return put_int_to_cc(dev, "D", y);
}


int get_box_x0(int dev, int *x)
{
	return get_int_from_cc(dev, "E", x);
}
		

int set_box_x0(int dev, int x)
{
	return put_int_to_cc(dev, "E", x);
}


int get_box_y_len(int dev, int *y)
{
	return get_int_from_cc(dev, "F", y);
}
		

int set_box_y_len(int dev, int y)
{
	return put_int_to_cc(dev, "F", y);
}


int get_box_x_len(int dev, int *x)
{
	return get_int_from_cc(dev, "G", x);
}
		

int set_box_x_len(int dev, int x)
{
	return put_int_to_cc(dev, "G", x);
}


int get_zoom_factor(int dev, int *z)
{
	return get_int_from_cc(dev, "H", z);
}
		

int set_zoom_factor(int dev, int z)
{
	return put_int_to_cc(dev, "H", z);
}


int get_auto_win(int dev, int *w)
{
	return get_int_from_cc(dev, "I", w);
}

		
/* This function turns the automatic window calculation function off
   and on */
int set_auto_win(int dev, int w)
{
	return put_int_to_cc(dev, "I", w);
}


int get_date_time(int dev, time_t *time)
{
	PVERB(PWARN,"get_date_time: not implemented\n");
	return -1; /* FIXME */
}
		

int set_date_time(int dev, time_t time)
{
	struct tm *t;
	char date_str[8];

	t = localtime(&time);
	sprintf(date_str,"%02d%02d%02d", 
		t->tm_mon + 1, t->tm_mday, t->tm_year + 1900 - 2000); 
	if (put_str_to_cc(dev, "L", date_str) != 0)
		return -1;

	sprintf(date_str,"%02d%02d%02d", t->tm_hour, t->tm_min, t->tm_sec);
	if (put_str_to_cc(dev, "M", date_str) != 0)
		return -1;

	return 0;
}


